require "action_notifier/message_delivery"
require "action_notifier/rescuable"

module ActionNotifier
  class Base < AbstractController::Base
    include ActiveSupport::Configurable
    include AbstractController::Rendering

    include AbstractController::Helpers
    include AbstractController::Translation
    include AbstractController::Callbacks
    include AbstractController::Caching
    include AbstractController::UrlFor
    include Rescuable

    class_attribute :url

    class_attribute :delivery_method
    self.delivery_method = :post_request_deliverer

    class << self
      def to(url)
        self.url = url
      end

      def notifier_name
        @notifier_name ||= anonymous? ? "anonymous" : name.underscore
      end

      attr_writer :notifier_name
      alias :controller_path :notifier_name

      # Aliased by ::default_options=
      def default(value = nil)
        self.default_params = default_params.merge(value).freeze if value
        default_params
      end
      # Allows to set defaults through app configuration:
      #
      #    config.action_mailer.default_options = { from: "no-reply@example.org" }
      alias :default_options= :default

      # Wraps an email delivery inside of <tt>ActiveSupport::Notifications</tt> instrumentation.
      #
      # This method is actually called by the <tt>Mail::Message</tt> object itself
      # through a callback when you call <tt>:deliver</tt> on the <tt>Mail::Message</tt>,
      # calling +deliver_mail+ directly and passing a <tt>Mail::Message</tt> will do
      # nothing except tell the logger you sent the email.
      def deliver_notification(notification) #:nodoc:
        ActiveSupport::Notifications.instrument("deliver.action_notifier") do |payload|
          set_payload_for_mail(payload, notification)
          yield # Let Mail do the delivery actions
        end
      end

    private

      def set_payload_for_mail(payload, mail)
        payload[:mailer]     = name
        payload[:message_id] = mail.message_id
        payload[:subject]    = mail.subject
        payload[:to]         = mail.to
        payload[:from]       = mail.from
        payload[:bcc]        = mail.bcc if mail.bcc.present?
        payload[:cc]         = mail.cc  if mail.cc.present?
        payload[:date]       = mail.date
        payload[:mail]       = mail.encoded
      end

      def method_missing(method_name, *args)
        if action_methods.include?(method_name.to_s)
          MessageDelivery.new(self, method_name, *args)
        else
          super
        end
      end

      def respond_to_missing?(method, include_all = false)
        action_methods.include?(method.to_s) || super
      end
    end

    attr_reader :message

    def initialize
      @message = Message.new
    end

    def notify(message:, url: nil, options: nil)
      @message.body = message
      @message.delivery_method = self.class.delivery_method
      @message.url = url || self.class.url
      @message.options = options || {}
      @message
    end

    def process(method_name, *args) #:nodoc:
      payload = {
        notifier: self.class.name,
        action: method_name,
        args: args,
        delivery_method: self.class.delivery_method
      }

      ActiveSupport::Notifications.instrument("process.action_notifier", payload) do
        super
      end
    end

    def notifier_name
      self.class.notifier_name
    end
  end
end
